"""
Simplifies URL parsing and manipulation tasks.
"""

import posixpath

from email.message import Message
from urllib.parse import urlparse, urlunparse, urlencode, parse_qs


class URL(object):
    """
    A helper object to make it easier to work with URLs and create endpoints.
    """

    __slots__ = ("scheme", "netloc", "path", "params", "query", "fragment")

    @classmethod
    def parse(cls, url: str, scheme: str = "", allow_fragments: bool = True) -> "URL":
        parsed = urlparse(url, scheme, allow_fragments)
        return cls(*parsed)

    def __init__(self, scheme, netloc, path, params, query, fragment):
        self.scheme = scheme
        self.netloc = netloc
        self.path = path
        self.params = params
        self.query = query
        self.fragment = fragment

    def resolve(self, *endpoint: str, query: dict = None) -> "URL":
        """
        Resolve the specified endpoint against this URL, returning a new URL object.

        The endpoint is joined with the path of this URL. If any component is an
        absolute path then all previous path components will be discarded.

        If a query is specified, it updates the query parameters of this URL and is
        encoded as a query string on the new URL.
        """
        path = posixpath.join(self.path, *endpoint)
        qs = self.parse_query()
        qs.update(query or {})

        return URL(
            self.scheme,
            self.netloc,
            path,
            self.params,
            urlencode(qs, doseq=True) if qs else "",
            self.fragment,
        )

    def parse_query(self) -> dict[str, list[str]]:
        """
        Parse the query string of this URL into a dictionary.
        """
        if self.query:
            return parse_qs(self.query)
        return {}

    def __str__(self):
        return urlunparse(self)

    def __iter__(self):
        for name in self.__slots__:
            yield getattr(self, name)


def parse_host(url: str) -> str:
    """
    Helper function to extract the host from a URL string.
    """
    parsed = urlparse(url, scheme="https", allow_fragments=False)
    if parsed.netloc:
        return parsed.netloc

    # if a domain is specified then it will be in the path
    return parsed.path.split("/")[0]


def parse_content_type(mime: str) -> tuple[str, dict[str, str]]:
    """
    Helper function to extract the main content type and parameters from a MIME string.
    """
    msg = Message()
    msg["content-type"] = mime
    params = msg.get_params()
    return params[0][0], dict(params[1:])



"""
Exception class hierarchy for guidelight related exceptions.
"""


class GuidelightError(Exception):
    """
    Base class for all guidelight related exceptions.
    """


class EndeavorError(GuidelightError):
    """
    Base class for all endeavor related exceptions.
    """


class ServerError(EndeavorError):
    """
    An error that is raised when the server returns a 500 status code.
    """


class ClientError(EndeavorError):
    """
    An unspecified client side error when the server returns a 400 status code.
    """


class AuthenticationError(ClientError):
    """
    Occurs when the server returns a 401 or 403 error code.
    """


class NotFound(ClientError):
    """
    Occurs when the server returns a 404 error code.
    """


class ValidationError(ClientError):
    """
    The payload is invalid or a preflight check has failed.
    """


class ReadOnlyEndpoint(EndeavorError):
    """
    The associated resource does not allow create, update, or delete methods
    """



"""
Manages JWT credentials that are returned from the Quarterdeck server on authentication.
"""

import jwt

from calendar import timegm
from datetime import datetime, timezone


class Credentials(object):
    """
    Credentials are composed of an access and a refresh token and are used to maintain
    authenticated state to the Quarterdeck server. The credentials also contains other
    claims such as the permissions allowed by the specified API key.
    """

    def __init__(self, access_token: str, refresh_token: str):
        self.access_token = Token(access_token)
        self.refresh_token = Token(refresh_token)

    def is_authenticated(self) -> bool:
        """
        Returns True if the access token is not expired
        """
        return not self.access_token.is_expired()

    def is_refreshable(self) -> bool:
        """
        Returns true if the refresh token is after the not before date and not expired.
        """
        return (
            not self.refresh_token.is_not_before()
            and not self.refresh_token.is_expired()
        )


class Token(object):
    """
    A token is a lightweight wrapper for JWT tokens and provides unverified decoding.
    """

    def __init__(self, token):
        self.token = token
        self._header = None
        self._claims = None

    def headers(self) -> dict:
        if self._header is None:
            self._header = jwt.get_unverified_header(self.token)
        return self._header

    def claims(self) -> dict:
        if self._claims is None:
            self._claims = jwt.decode(self.token, options={"verify_signature": False})
        return self._claims

    def is_expired(self) -> bool:
        """
        Returns true if the current time is after the exp claim.
        """
        exp = self.claims().get("exp", None)
        if exp is None:
            return True

        return timegm(datetime.now(tz=timezone.utc).utctimetuple()) > exp

    def is_not_before(self) -> bool:
        """
        Returns true if the current time is before the nbf claim.
        """
        nbf = self.claims().get("nbf", None)
        if nbf is None:
            return True

        return timegm(datetime.now(tz=timezone.utc).utctimetuple()) < nbf

    def __str__(self) -> str:
        return self.token



"""
A library for developing task-oriented AI systems that integrate with Endeavor.
"""

##########################################################################
## Module Info
##########################################################################

# Import the version number at the top level
from .version import get_version, __version_info__


##########################################################################
## Package Version
##########################################################################

__version__ = get_version(short=True)


##########################################################################
## Primary API Entry Point
##########################################################################

from dotenv import load_dotenv
from .client import Client


def connect(url=None, client_id=None, client_secret=None, timeout=None):
    """
    Create an API client with the specified URL and api key material. If not specified,
    this function will first load any .env files in the local path, then attempt to
    configure the client from the environment.

    Parameters
    ----------
    url : str
        The URL of your Endeavor server (e.g. https://guidelight.dev). If not
        set, it is discovered from the $ENDEAVOR_URL environment variable.

    client_id : str
        The Client ID from your API Key to access your Endeavor server. If not set, it
        is discovered from the $ENDEAVOR_CLIENT_ID environment variable.

    client_secret : str
        The Client Secret from your API Key to access your Endeavor server. If not set,
        it is discovered from the $ENDEAVOR_CLIENT_SECRET environment variable.

    timeout : float
        The number of seconds to wait for a response until error.
    """

    if url is None or client_id is None or client_secret is None:
        # We need to load information from the environment
        load_dotenv()

    # create the client and perform the pre-flight now to authorize the client
    # now so the client's first actual data request isn't delayed by seconds
    client = Client(
        url=url, client_id=client_id, client_secret=client_secret, timeout=timeout
    )
    client._pre_flight(require_authentication=True)
    return client



from guidelight.client import Client
from pytest_httpserver import HTTPServer


def test_preflight():
    """
    Test preflight authentication happens implicitly.
    """

    # Create a test server
    server = HTTPServer()
    server.start()

    # Expect authentication request
    server.expect_request("/v1/authenticate").respond_with_json(
        {
            "access_token": "access",
            "refresh_token": "refresh",
        }
    )

    # Start a guidelight client
    client = Client(
        "http://localhost:%d" % server.port, client_id="id", client_secret="secret"
    )

    # Perform auth required request to trigger the preflight code
    server.expect_request(
        "/v1/authrequired",
        headers={"Authorization": "Bearer access"},
    ).respond_with_json([])
    client.post({}, "/v1/authrequired")



import pytest

from guidelight.url import URL, parse_host, parse_content_type


@pytest.mark.parametrize(
    "url",
    [
        "http://example.com",
        "https://example.com/path/to/resource?query=param#fragment",
        "ftp://ftp.example.com/resource",
        "http://example.com:8080",
        "http://localhost",
        "http://localhost:9000/path",
        "http://example.local",
    ],
)
def test_url_parse(url):
    parsed = URL.parse(url)
    assert str(parsed) == url


@pytest.mark.parametrize(
    "url,endpoint,query,expected",
    [
        (
            "https://guidelight.dev",
            ("v1", "agents"),
            None,
            "https://guidelight.dev/v1/agents",
        ),
        (
            "https://guidelight.dev/v1",
            ("agents",),
            None,
            "https://guidelight.dev/v1/agents",
        ),
        (
            "https://guidelight.dev/v1",
            ("/", "v2", "agents"),
            None,
            "https://guidelight.dev/v2/agents",
        ),
        (
            "https://guidelight.dev?ordering=created",
            ("v1", "agents"),
            None,
            "https://guidelight.dev/v1/agents?ordering=created",
        ),
        (
            "https://guidelight.dev?ordering=created",
            ("v1", "agents"),
            {"ordering": "modified"},
            "https://guidelight.dev/v1/agents?ordering=modified",
        ),
        (
            "https://guidelight.dev?ordering=created",
            ("v1", "agents"),
            {"include": "archives"},
            "https://guidelight.dev/v1/agents?ordering=created&include=archives",
        ),
    ],
)
def test_resolve(url, endpoint, query, expected):
    base = URL.parse(url)
    resolved = base.resolve(*endpoint, query=query)
    assert str(resolved) == expected


@pytest.mark.parametrize("input,expected", [
    ("http://example.com", "example.com"),
    ("https://example.com/path", "example.com"),
    ("ftp://ftp.example.com/resource", "ftp.example.com"),
    ("http://example.com:8080", "example.com:8080"),
    ("http://localhost", "localhost"),
    ("http://localhost:9000/path", "localhost:9000"),
    ("http://example.local", "example.local"),
])
def test_parse_host(input, expected):
    actual = parse_host(input)
    assert actual == expected


@pytest.mark.parametrize("input,mime,params", [
    ("text/html; charset=UTF-8", "text/html", {"charset": "UTF-8"}),
    ("application/json", "application/json", {}),
])
def test_parse_content_type(input, mime, params):
    result_mime, result_params = parse_content_type(input)
    assert result_mime == mime
    assert result_params == params



"""
Implements the API client that manages authentication and cookie state of requests to
and from an Endeavor server. These requests integrate guidelight with Endeavor for AI
projects and training.
"""

import os
import logging

from requests import Response
from platform import python_version
from requests.sessions import Session
from requests.adapters import HTTPAdapter

from .version import get_version
from .credentials import Credentials
from .url import URL, parse_content_type
from .exceptions import ClientError, ServerError
from .exceptions import AuthenticationError, NotFound

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


# Setup debug logging for guidelight
logger = logging.getLogger("endeavor")


# Environment variables for local configuration
ENV_URL = "ENDEAVOR_URL"
ENV_CLIENT_ID = "ENDEAVOR_CLIENT_ID"
ENV_CLIENT_SECRET = "ENDEAVOR_CLIENT_SECRET"

# Default header values
ACCEPT = "application/json"
ACCEPT_LANG = "en-US,en"
ACCEPT_ENCODE = "gzip, deflate, br"
CONTENT_TYPE = "application/json; charset=utf-8"


class Client(object):
    """
    Client manages the connection/session to an Endeavor server and makes API requests.

    Parameters
    ----------
    url : str, optional
        The base URL of the Endeavor server. If not provided, defaults to the value of
        the `ENDEAVOR_URL` environment variable. If neither is set, an error is raised.

    client_id : str, optional
        The client ID of your API key for authentication. If not provided, defaults to
        the value of the `ENDEAVOR_CLIENT_ID` environment variable. If neither is set,
        an error is raised.

    client_secret : str, optional
        The client secret of your API key for authentication. If not provided, defaults
        to the value of the `ENDEAVOR_CLIENT_SECRET` environment variable. If neither is
        set, an error is raised.

    timeout : float, optional
        The number of seconds to wait for a response until error.

    pool_connections : int, default=8
        The number of urllib3 connections to cache in a pool.

    pool_maxsize : int, default=16
        The maximum number of connections to save in the pool.

    max_retries : int, default=3
        The maximum number of retries for a request. Note, this only applies to failed
        DNS lookups, socket connections and connection timeouts, never to requests where
        data has made it to the server.
    """

    def __init__(
        self,
        url=None,
        client_id=None,
        client_secret=None,
        timeout=None,
        pool_connections=8,
        pool_maxsize=16,
        max_retries=3,
    ):
        self._host = None
        self._creds = None
        self._prefix = None

        self.url = url or os.environ.get(ENV_URL, "")
        self.client_id = client_id or os.environ.get(ENV_CLIENT_ID, None)
        self.client_secret = client_secret or os.environ.get(ENV_CLIENT_SECRET, None)

        user_agent = f"guidelight/{get_version(short=True)} python/{python_version()}"
        self._headers = {
            "Accept": ACCEPT,
            "Accept-Language": ACCEPT_LANG,
            "Accept-Encoding": ACCEPT_ENCODE,
            "Content-Type": CONTENT_TYPE,
            "User-Agent": user_agent,
        }

        # Configure HTTP requests with the requests library
        self.timeout = timeout
        self.session = Session()
        self.adapter = HTTPAdapter(
            pool_connections=pool_connections,
            pool_maxsize=pool_maxsize,
            max_retries=max_retries,
        )
        self.session.mount(self.prefix + "://", self.adapter)

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        if value is None:
            self._timeout = (10.0, 30.0)
        else:
            self._timeout = value

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._host = None
        self._prefix = None
        self._url = URL.parse(value) if value else None

    @property
    def host(self):
        if self._host is None and self.url:
            parsed = self.url
            if parsed.netloc:
                self._host = parsed.netloc
            else:
                # if a domain is specified then it will be in the path
                self._host = parsed.path.split("/")[0]

        return self._host

    @property
    def prefix(self):
        if self._prefix is None:
            if not self.host:
                raise ValueError("cannot compute prefix without host")

            if self.is_localhost():
                self._prefix = "http"
            else:
                self._prefix = "https"

        return self._prefix

    def get(
        self,
        *endpoint: tuple[str],
        query: dict = None,
        require_authentication: bool = True,
    ) -> dict:
        headers = self._pre_flight(require_authentication=require_authentication)
        url = self._make_endpoint(*endpoint, query=query)

        logger.debug(f"GET {repr(url)}")

        rep = self.session.get(str(url), headers=headers, timeout=self.timeout)

        return self.handle(rep)

    def post(
        self,
        data,
        *endpoint: tuple[str],
        query: dict = None,
        require_authentication: bool = True,
    ) -> dict:
        headers = self._pre_flight(require_authentication=require_authentication)
        url = self._make_endpoint(*endpoint, query=query)

        logger.debug(f"POST {repr(url)}")

        rep = self.session.post(
            str(url), json=data, headers=headers, timeout=self.timeout
        )

        return self.handle(rep)

    def put(
        self,
        data,
        *endpoint: tuple[str],
        query: dict = None,
        require_authentication: bool = True,
    ) -> dict:
        headers = self._pre_flight(require_authentication=require_authentication)
        url = self._make_endpoint(*endpoint, query=query)

        logger.debug(f"PUT {repr(url)}")

        rep = self.session.put(
            str(url), json=data, headers=headers, timeout=self.timeout
        )

        return self.handle(rep)

    def delete(
        self,
        *endpoint: tuple[str],
        query: dict = None,
        require_authentication: bool = True,
    ) -> dict:
        headers = self._pre_flight(require_authentication=require_authentication)
        url = self._make_endpoint(*endpoint, query=query)

        logger.debug(f"DELETE {repr(url)}")

        rep = self.session.delete(str(url), headers=headers, timeout=self.timeout)

        return self.handle(rep)

    def handle(self, rep: Response) -> dict:
        """
        Handle the response from an API request, raising an error if the request failed.
        """
        if rep.status_code == 401 or rep.status_code == 403:
            raise AuthenticationError("authentication failed")

        elif rep.status_code == 204:
            return None

        elif 200 <= rep.status_code < 300:
            mimetype, _ = parse_content_type(rep.headers.get("Content-Type"))
            if mimetype == "application/json":
                return rep.json()
            else:
                return rep.content

        elif 400 <= rep.status_code < 500:
            logger.warning(f"client error: {rep.status_code} {repr(rep.content)}")
            message = f"{rep.status_code} response from {self.host}"

            try:
                err = rep.json()
                if "error" in err:
                    message = err["error"]
                if "errors" in err:
                    message += ":\n  " + "\n  ".join(
                        [f"{e['field']}: {e['error']}" for e in err["errors"]]
                    )
            except JSONDecodeError:
                pass

            if rep.status_code == 404:
                raise NotFound(message)
            else:
                raise ClientError(message)

        elif 500 <= rep.status_code < 600:
            logger.warning(f"server error: {rep.status_code} {repr(rep.content)}")
            message = f"{rep.status_code} response from {self._host}]"

            try:
                err = rep.json()
                if "error" in err:
                    message = err["error"]
            except JSONDecodeError:
                pass

            raise ServerError(message)

        else:
            raise ValueError(f"unhandled status code {rep.status_code}")

    def is_authenticated(self) -> bool:
        """
        Returns True if there are JWT claims with a valid access token
        """
        return self._creds is not None and self._creds.is_authenticated()

    def is_refreshable(self) -> bool:
        """
        Returns True if there are JWT claims with a valid refresh token
        """
        return self._creds is not None and self._creds.is_refreshable()

    def is_localhost(self) -> bool:
        """
        Returns true if the host is a local domain (e.g. localhost)
        """
        host = self.host
        if ":" in host:
            host = host.split(":")[0]
        return host == "localhost" or host.endswith(".local")

    def _make_endpoint(self, *endpoint: tuple[str], query: dict = None) -> URL:
        return self.url.resolve("/", "v1", *endpoint, query=query)

    def _pre_flight(self, require_authentication: bool = True) -> dict[str, str]:
        if not self.url:
            raise ClientError("no Endeavor URL has been configured")

        request_headers = {}
        request_headers.update(self._headers)

        if require_authentication:
            request_headers.update(self._authentication_headers())
        return request_headers

    def _authentication_headers(self) -> dict[str, str]:
        if not self.is_authenticated():
            # We need to reauthenticate, determin if we can refresh our credentials
            if self.is_refreshable():
                self._creds = self._reauthenticate()
            else:
                self._creds = self._authenticate()

        return {"Authorization": "Bearer " + str(self._creds.access_token)}

    def _authenticate(self) -> Credentials:
        if not self.client_id or not self.client_secret:
            raise AuthenticationError("no client id or secret specified")

        apikey = {"client_id": self.client_id, "client_secret": self.client_secret}
        rep = self.post(apikey, "authenticate", require_authentication=False)
        return Credentials(rep["access_token"], rep["refresh_token"])

    def _reauthenticate(self) -> Credentials:
        if not self._creds.refresh_token:
            raise AuthenticationError("no refresh token available")

        refresh = {"refresh_token": str(self._creds.refresh_token)}
        rep = self.post(refresh, "reauthenticate", require_authentication=False)
        return Credentials(rep["access_token"], rep["refresh_token"])
